#include "mad.h"
#include "speechy.h"
#include "voice_mgr.h"
#include <time.h>
#include <sys/stat.h>

static const char *load_if_changed(
	const char *file,
	unsigned int *lasttime,
	char *buf,
	unsigned int len)
{
	struct stat b;

	if(!stat(file, &b))
	{
		unsigned int changetime = b.st_mtime;

		if(changetime != *lasttime)
		{
			FILE *fp;
			unsigned int size;

			fp = fopen(file, "r+");

			size = fread(buf, 1, len, fp);

			PRINTF("reloaded '%s' (%u vs %u) %d bytes\n",
				file,
				changetime,
				*lasttime,
				size);

			fclose(fp);

			*lasttime = changetime;
			buf[size] = 0;

			return buf;
		}
	}

	return NULL;
}


// TODO: Scan for extra commands at the start
static const char *scan_commands(const char *text)
{
	const char *ptr = text;

	return ptr;
}


const char *speech_file = NULL;
static VOICE_MGR *voice = NULL;

static int16_t mixbuf[4096];

static void speech_callback(
	int16_t *buffer,
	int stereo_samples,
	int rate)
{
	int i;
	int16_t *ptr = buffer;
	int read;
	float freqscale = rate / 44100.0;

	read = speechy_read(voice->voice, ptr, stereo_samples);
	//~ read = speechy_read(voice->voice, mixbuf, stereo_samples);

	//~ for(i = 0; i < read; i++)
	//~ {
		//~ // mixbuf[i] *= SIN(i * freqscale * 16);
		//~ *ptr++ = mixbuf[i];
		//~ *ptr++ = mixbuf[i];
	//~ }
//~ 
	//~ for(i = 0; i < stereo_samples - read; i++)
	//~ {
		//~ *buffer++ = 0;
	//~ }
}


void game_reload_changed_file(void)
{
	static unsigned int lasttime = 0;
	static char buf[1024];
	const char *text;
	short codes[512];
	int length;

	if(mad_ticks & 15) return;

	text = load_if_changed(speech_file, &lasttime, buf, 1024);

	if(!text) return;

	text = scan_commands(text);

	// TODO: Make speech here!
	if(!voice_parse(voice, text, codes, 512)) return;

	speechy_reset(voice->voice);
	length = speechy_sequence(voice->voice, codes);

	PRINTF("SPEECH: %s\n", text);
	PRINTF("length: %d samples\n", length);
}


static void update(void)
{
	game_reload_changed_file();

	//~ if(!(mad_ticks & 63))
	//~ {
		//~ int i = rnd(0, voice->phon_count - 1);
		//~ WAV *wav = voice->phon_list[i].sample;
//~ 
		//~ PRINTF("allophone '%s' %p\n",
			//~ voice->phon_list[i].name,
			//~ wav);
//~ 
		//~ wave_just_play(wav);
	//~ }
}


static void enter(void)
{
	mad_set_glitch_callback(speech_callback);
}


STATE main_state = 
{
	enter,
	update,
	NULL,
	NULL,
};


static void setup_audio(unsigned int rate)
{
	mad_init_audio_stream(rate, 0);
	mad_enable_wave(1);

	voice = voice_load("allophones/sam.txt");
}


static int app_init(int argc, char *argv[])
{
	unsigned int rate = 0;

	MADFUNCTAG;

	PRINTF("argc %d\n", argc);
	if(argc > 1)
	{
		speech_file = argv[1];
	}

	if(argc > 2)
	{
		sscanf(argv[2], "%u", &rate);
	}

	mad_set_logic_rate(60);

	if(!rate) rate = 22050;

	PRINTF("speech_file: '%s' rate: %u\n", speech_file, rate);

	setup_audio(rate);

 	state_switch(&main_state);

	return 0;
}


static void app_exit(void)
{
	MADFUNCTAG;
	voice_unload(voice);
}


int main(int argc, char *argv[])
{
	mad_init();

	mad_set_app_name("speechy");

	app_init(argc, argv);

	mad_run(0);

	app_exit();

	mad_exit();

	return 0;
}
