/*
	TODO:
*/

#include <stdlib.h>
#include <string.h>
#include "speechy.h"

#define SILENCE			0
#define DEF_SAMPLE_RATE	8000
#define MILLISECONDS	1000

//~ #define BEEP_HIGH		192
//~ #define BEEP_LOW		64

#define BEEP_HIGH		16384
#define BEEP_LOW		-16384

#define BEEP_LEN		500
#define BEEP_FREQ1		240
#define BEEP_FREQ2		120

// Initialize a new voice with "count" amount of allophone slots
SPEECHY_VOICE *speechy_init(int num, int freq)
{
	SPEECHY_VOICE *voice;
	SPEECHY_ALLOPHONE *table;

	if(num > END_SEQ) return NULL;

	voice=(SPEECHY_VOICE*)malloc(sizeof(SPEECHY_VOICE));

	if(!voice) return NULL;

	table=(SPEECHY_ALLOPHONE*)calloc(sizeof(SPEECHY_ALLOPHONE), num);

	if(!table)
	{
		free(voice);
		return NULL;
	}

	if(freq == 0)
	{
		freq = DEF_SAMPLE_RATE;
	}

	voice->max  =num;
	voice->count = 0;
	voice->sequence = NULL;
	voice->pos = 0;
	voice->freq = freq;

	voice->table = table;

	return voice;
}


int speechy_add(SPEECHY_VOICE *voice, int16_t *buffer, int length)
{
	SPEECHY_ALLOPHONE *allophone;

	// Make sure we have a slot left
	if(voice->count >= voice->max) return 0;

	allophone = &voice->table[voice->count];

	if(!buffer || !length)
	{
		allophone->data = NULL;
		allophone->length = voice->freq >> 2;
	}
	else
	{
		allophone->data = buffer;
		allophone->length = length;
	}

	// Increment allophone count
	voice->count++;

	return length;
}


int speechy_add_silence(SPEECHY_VOICE *voice, int t)
{
	SPEECHY_ALLOPHONE *allophone;
	unsigned int length;

	MADFUNCTAG;

	// Make sure we have a slot left
	if(voice->count >= voice->max) return 0;

	if(!t) return 0;

	allophone = &voice->table[voice->count];

	// Calculate the number of samples in "time" milliseconds
	length = (voice->freq / MILLISECONDS) * t;

	// No sound, no data
	allophone->data = NULL;
	allophone->length = length;

	// Increment allophone count
	voice->count++;

	return length;
}


int speechy_sequence(SPEECHY_VOICE *voice, short *sequence)
{
	int16_t *ptr;
	int len;

	if(!sequence)
	{
		voice->sequence = NULL;
		voice->pos = 0;
		return 0;
	}

	ptr = sequence;
	len = 0;

	// Get the total length of the sequence in samples
	while(*ptr != END_SEQ)
	{
		// TODO: if ERR_SEQ modify sequence
		if(*ptr >= voice->count ||
			*ptr == ERR_SEQ ||
			!voice->table[*ptr].data)
		{
			len += voice->freq >> 2;
		}
		else
		{
			len += voice->table[*ptr].length;
		}

		ptr++;
	}

	voice->sequence = sequence;
	voice->pos = 0;

	return len;
}


static int16_t process_sample(int16_t sample, int pos, int freq)
{
	return sample * SIN(pos * freq / 44100.0 * 6);
}


int speechy_read(SPEECHY_VOICE *voice, int16_t *buffer, int length)
{
	int i, pos, read_count = 0, current, next;
	int len = 0, mod;
	short *ptr;
	int16_t *out = buffer;
	int16_t *in;
	int16_t sample;
	int16_t a, b, sum;
	int j;
	float freqscale = voice->freq / 44100.0;

	if(!voice->sequence || !length || !buffer) return 0;

	// TODO: if ERR_SEQ then stream out error beep
	// Set positions to our next sample in the sequence
	ptr = voice->sequence;
	pos = voice->pos;
	current = ptr[0];

	in = voice->table[current].data;

	next = 1;

	if(current != END_SEQ)
	{
		// Give required samples to the user
		for(i = 0; i < length; i++)
		{
			// Error beep
			if(current == ERR_SEQ)
			{
				if(next)
				{
					len = voice->freq >> 2;
					next = 0;
				}

				if(pos % len < len / 2)
				{
					mod = voice->freq / BEEP_FREQ1;
				}
				else
				{
					mod = voice->freq / BEEP_FREQ2;
				}

				sample = (pos % mod < (mod >> 1)) ? BEEP_HIGH : BEEP_LOW;
			}
			// Allophone
			else if(in)
			{
				if(next)
				{
					len = voice->table[current].length;
					next = 0;
				}

				sample = in[pos];
				sample = process_sample(sample, pos, voice->freq);
			}
			// Silence allophone
			else
			{
				if(next)
				{
					len = voice->table[current].length;
					next = 0;
				}

				sample = 0;
			}

			for(j = 0; j < 2; j++)
			{
				a = *out;
				sum = a + sample;

				// signed overflow check
				if(((a ^ sum) & (sample ^ sum)) < 0)
				{
					// clip
					sum = (sum < 0) ? 32767 : -32768;
				}

				*out++ = sum;
			}
//~ //~
			//~ *out++ = sample;

			pos++;

			// We've hit the end of this allophone, go to the next in sequence
			if(pos >= len)
			{
				pos = 0;
				ptr++;
				current = *ptr;

				//~ PRINTF("CURRENT %d\n", current);

				if(current == END_SEQ)
				{
					in = NULL;
					i++;
					break;
				}

				if(current != ERR_SEQ)
				{
					in = voice->table[current].data;
				}
				else
				{
					in = NULL;
				}

				next = 1;
			}
		}

		read_count = i;
	}
	else
	{
		// Already at the end, so we read nothing
		read_count = 0;
	}

	// End of the sequence, so reset everything
	if(*ptr == END_SEQ)
	{
		ptr = NULL;
		pos = 0;
	}

	// Adjust the voice place-keepers before exiting
	voice->sequence = ptr;
	voice->pos = pos;

	return read_count;
}


void speechy_reset(SPEECHY_VOICE *voice)
{
	voice->sequence = NULL;
	voice->pos = 0;
}


void free_allophone(SPEECHY_VOICE *voice, int i)
{
	//~ SPEECHY_ALLOPHONE *allophone;
//~ 
	//~ allophone = &voice->table[index];

	//~ if(allophone && allophone->data && allophone->copy)
	//~ {
		//~ free(allophone->data);
	//~ }
}


void speechy_exit(SPEECHY_VOICE *voice)
{
	// Free all the allophones
	while(voice->count)
	{
		free_allophone(voice, voice->count);
		voice->count--;
	}

	free(voice->table);
	free(voice);
}
